# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from playwright.async_api import Page
from tests.server import Server


async def test_should_work(page: Page, server: Server) -> None:
    await page.evaluate(
        """() => {
        globalThis.objectToDestroy = { hello: 'world' };
        globalThis.weakRef = new WeakRef(globalThis.objectToDestroy);
    }"""
    )
    await page.request_gc()
    assert await page.evaluate("() => globalThis.weakRef.deref()") == {"hello": "world"}

    await page.request_gc()
    assert await page.evaluate("() => globalThis.weakRef.deref()") == {"hello": "world"}

    await page.evaluate("() => globalThis.objectToDestroy = null")
    await page.request_gc()
    assert await page.evaluate("() => globalThis.weakRef.deref()") is None
